#!/bin/env python
# vim:fileencoding=utf-8:ft=python
#
# Desktop widgets implementation in python
# Following k.i.s.s.
# Copyright (c) 2007, Phillip Berndt
#
# Version: 0.1
#
import sys
try:
	import gtk, cairo, gobject, os, getopt
except ImportError:
	print >> sys.stderr, "FAILED TO IMPORT A REQUIRED DEPENDENCY"
	error = str(sys.exc_info()[1])
	if error[0:16] == "No module named ":
		print >> sys.stderr, "Please install python module", error[16:]
	else:
		print >> sys.stderr, error
	print >> sys.stderr
	sys.exit(1)

PLUGIN_DIRECTORY = os.path.join(os.path.dirname(sys.argv[0]), "plugins/")
CONFIG_DIRECTORY = "~/.config/widgets/"
DISABLED_PLUGINS_FILENAME = os.path.expanduser(CONFIG_DIRECTORY + "disabled")

__doc__ = \
"""

	Desktop widgets implementation in python

	Documentation for users:

	Call ./widgets.py to run this program. It will NOT fork itself into 
	background. Delete/Copy widgets into the plugins/ folder to (de)activate
	them. (Or use the new configuration item in the popup menu of each plugin)
	The only place outside of the application directory, where this program
	writes to, is ~/.config/widgets/.

	This program requires the composite extension.


	Documentation for plugin developers:

	Developing plugins / widgets is quite simple. Create a python file and
	put it into the plugins/ directory. This script will load it upon
	initialisation and supply it with some usefull attributes:

	window:			A transparent, draggable gtk.Window which is bottommost
				You usually don't want to use this
	widget:			An eventbox inside this window. Put your widgets inside
				of it.  If you need user interactivity this might lead
				to a problem - the eventbox is necessary for the moving
				and popup stuff. Widgets shouldn't require too much
				interactivity anyway ;)
	menu:			A gtk.menu which is used as a popup menu. Put your
				menu stuff here
	settingsDirectory:	A string containing a path where you can save your
				widget's configuration. Don't use the name "position",
				it's used internally.
	pluginDirectory:	A string containing the directory where plugins are
				saved in. You are free to create subdirectories etc.
	
	You can create functions which will be called by this script:

	initialize:		Called upon program start. Build your widget here.
				It's YOUR part to call window.show_all()!


	Any ideas or improvements?
	-> mail at pberndt dot com
"""
def configure(*param):
	"""
		Show a configuration dialog for this program
	"""
	global disabledPlugins
	global plugins

	dialog = gtk.Dialog("Configure widgets")
	endisableVbox = gtk.VBox()
	dialog.vbox.add(endisableVbox)
	endisableVbox.add(gtk.Label("Enable the following plugins:"))
	endisableList = {}
	for plugin in listPlugins():
		checkbox = gtk.CheckButton(plugin)
		checkbox.set_active(plugin not in disabledPlugins)
		endisableVbox.add(checkbox)
		endisableList[plugin] = checkbox
	applyButton = gtk.Button(stock='gtk-apply')
	def applySave(*param):
		global disabledPlugins
		disabledPlugins = [ plugin for plugin in endisableList if not endisableList[plugin].get_active() ]
		disabledPluginsFile = open(DISABLED_PLUGINS_FILENAME, "w")
		disabledPluginsFile.write("\n".join(disabledPlugins))
		disabledPluginsFile.close()

		msg = gtk.MessageDialog(dialog, gtk.DIALOG_MODAL, gtk.MESSAGE_WARNING, gtk.BUTTONS_YES_NO,
			"Some of the settings require you to restart widgets before they take effect.\nRestart now?")
		msg.set_title("Widgets")
		response = msg.run()
		msg.destroy()
		if response == gtk.RESPONSE_YES:
			for plugin in plugins:
				plugin.window.destroy()
			os.execv(sys.argv[0], sys.argv)
			os.exit(1)
		dialog.destroy()
	applyButton.connect("clicked", applySave)
	cancelButton = gtk.Button(stock='gtk-cancel')
	cancelButton.connect("clicked", lambda *x: dialog.destroy())
	dialog.action_area.add(cancelButton)
	dialog.action_area.add(applyButton)
	dialog.set_modal(True)
	dialog.show_all()

def createWidgetWindow(title="Unnamed widget"):
	"""
		Create a (desktop-)widget window and return a reference on it
		The window has some special attributes:
		 - area, which is the container to be used by each widget
		 - menu, which a widget MAY use to implement popup menu items
	"""
	# Create window
	wnd = gtk.Window()
	wnd.set_decorated(False)
	wnd.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_SPLASHSCREEN)
	wnd.set_keep_below(True)
	wnd.set_title("Widget: " + title)
	wnd.stick()
	wnd.connect("delete-event", gtk.main_quit)
	# Give it an eventbox for movement etc. stuff
	eventBox = gtk.EventBox()
	eventBox.set_visible_window(False)
	wnd.add(eventBox)
	wnd.area = eventBox
	eventBox.move = False
	def beginMove(widget, event):
		if event.button == 1:
			# Left mousebutton = Move
			widget.move = True
			widget.last = event.get_root_coords()
		elif event.button == 2:
			# Middle one = Add decoration to window
			wnd.set_decorated(not wnd.get_decorated())
		elif event.button == 3:
			# Right one = Show popip menu
			wnd.menu.popup(None, None, None, 3, event.time)
	eventBox.connect("button-press-event", beginMove)
	def endMove(widget, event):
		widget.move = False
	eventBox.connect("button-release-event", endMove)
	def move(widget, event):
		if widget.move:
			movedBy = [ x[0] - x[1] for x in zip(event.get_root_coords(), widget.last) ]
			moveTo = [ int(sum(x)) for x in zip(wnd.get_position(), movedBy) ]
			apply(wnd.move, moveTo)
			widget.last = event.get_root_coords()
	eventBox.connect("motion-notify-event", move)
	# Create popup menu
	wnd.menu = gtk.Menu()
	quitItem = gtk.MenuItem("Close widget")
	quitItem.connect("activate", lambda *x: wnd.destroy())
	wnd.menu.add(quitItem)
	quitItem = gtk.MenuItem("Quit")
	def quitAll(widget):
		global plugins
		for plugin in plugins:
			plugin.window.destroy()
		gtk.main_quit()
	quitItem.connect("activate", quitAll)
	wnd.menu.add(quitItem)
	configItem = gtk.MenuItem("Configure widgets")
	configItem.connect("activate", configure)
	wnd.menu.add(configItem)
	wnd.menu.add(gtk.SeparatorMenuItem())
	wnd.menu.show_all()
	# Make window transparent
	try:
		wnd.set_app_paintable(True)
		wnd.set_colormap(wnd.get_screen().get_rgba_colormap())
		def expose(widget, event):
			context = widget.window.cairo_create()
			context.set_source_rgba(1.0, 1.0, 1.0, 0.0)
			context.set_operator(cairo.OPERATOR_SOURCE)
			context.paint()
			return False
		wnd.connect("expose-event", expose)
	except:
		print "Failed to initialize compositing. You won't have nice widgets..."
		# Copy background on exposé. This is not nice :/
		wnd.set_app_paintable(True)
		def expose(wnd, event):
			"""
				Copy the root window's background image into the
				widget's background
			"""
			screen = wnd.get_screen()
			root = screen.get_root_window()
			swidth, sheight = screen.get_width(), screen.get_height()
			width, height = wnd.get_size()
			bleft = btop = 0
			left, top = wnd.get_position()
			if left < 0:
				bleft = -left
				left = 0
			if top < 0:
				btop = -top
				top = 0
			if width + left > swidth: width = swidth - left
			if height + top > sheight: height = sheight - top

			bgData = root.property_get(gtk.gdk.atom_intern("_XROOTPMAP_ID"), gtk.gdk.TARGET_PIXMAP, False)
			bg = gtk.gdk.pixmap_foreign_new(bgData[2][0])
			bgPixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, False, 8, width, height)
			bgPixbuf.get_from_drawable(bg, gtk.gdk.colormap_get_system(), left, top, bleft, btop, width - bleft, height - btop)

			wnd.window.draw_pixbuf(None, bgPixbuf, 0, 0, 0, 0, width, height, gtk.gdk.RGB_DITHER_NONE, 0, 0)
			return False
		wnd.connect("expose-event", expose)
		def configureEvent(*param):
			wnd.queue_draw()
		wnd.connect("configure-event", configureEvent)

	return wnd

def listPlugins():
	"""
		Get a list of all plugins
	"""
	for plugin in os.listdir(PLUGIN_DIRECTORY):
		if plugin[-3:] != ".py":
			continue
		yield plugin

def call(plugin, function, *args):
	"""
		Check if plugin has an attribute named like function and call
		it with given args
	"""
	if function in dir(plugin):
		caller = getattr(plugin, function)
		apply(caller, args)

def call_async(*args):
	"""
		Like call, but asyncronous
	"""
	def thread():
		gtk.gdk.threads_enter()
		apply(call, args)
		gtk.gdk.threads_leave()
		return False
	gobject.idle_add(thread)

plugins = []
disabledPlugins = []
if __name__ == "__main__":
	gobject.threads_init()
	gtk.gdk.threads_init()
	gtk.gdk.threads_enter()
	if os.access(DISABLED_PLUGINS_FILENAME, os.F_OK):
		disabledPluginsFile = open(DISABLED_PLUGINS_FILENAME, "r")
		disabledPlugins = [ n.strip() for n in disabledPluginsFile.readlines() ]
		disabledPluginsFile.close()
	activePlugins = filter(lambda x: x not in disabledPlugins, listPlugins())
	for pluginName in activePlugins:
		# Create path
		settingsDirectory = os.path.expanduser(CONFIG_DIRECTORY + "%s/" % pluginName[:-3])
		path = ""
		for bit in settingsDirectory.split(os.path.sep):
			path += bit + "/"
			if not os.access(path, os.F_OK):
				os.mkdir(path)
		# Import plugin
		plugin = __import__("plugins/" + pluginName[:-3])
		plugin.settingsDirectory = settingsDirectory
		plugin.pluginDirectory = PLUGIN_DIRECTORY
		plugin.window = createWidgetWindow(pluginName)
		plugin.window.plugin = plugin
		plugin.widget = plugin.window.area
		plugin.menu = plugin.window.menu
		# Position widget
		plugin.window.positionFile = position = os.path.join(settingsDirectory, "position")
		if os.access(position, os.F_OK):
			try:
				file = open(position)
				data = [ int(x) for x in file.read().split(",", 2) ]
				assert len(data) == 2
				file.close()
				apply(plugin.window.move, data)
			except: 
				plugin.window.move((len(plugins) % 5) * 150, (len(plugins) / 5) * 150)
		else:
			plugin.window.move((len(plugins) % 5) * 150, (len(plugins) / 5) * 150)
		def backupPosition(widget, event):
			widget.last_known_position = widget.get_position()
		plugin.window.connect("configure-event", backupPosition)
		def savePositionOrQuit(widget):
			global plugins
			plugins.remove(widget.plugin)
			try:
				file = open(widget.positionFile, "w")
				file.write(",".join([ str(x) for x in widget.last_known_position ]))
				file.close()
			except: pass
			if len(plugins) == 0:
				gtk.main_quit()
		plugins.append(plugin)
		plugin.window.connect("hide", savePositionOrQuit)
		call_async(plugin, "initialize")
	if len(activePlugins) is 0:
		configure()
		os.exit(0)
	gtk.gdk.threads_leave()
	gtk.main()
